# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .error import BaseError, NotFunctionError, NotAsyncFunctionError

from .is_validators import (
    is_any_string,
    is_bytes,
    is_bytearray,
    is_string,
    is_coroutine,
    is_dict,
    is_enum,
    is_enum_subclass,
    is_float,
    is_function,
    is_async_function,
    is_int,
    is_iterable,
    is_list,
    is_number,
    is_none,
    is_not_none,
    is_path
)

from .validators import AbstractValidator, FunctionValidator, AsyncFunctionValidator
from .manager import IValidatorManager, ValidatorManager, ValidatorTypes, ValidatorList
from .api import is_validator_subclass, create, checker, check_function, check_async_function


# @Time     : 2021/4/13
# @Project  : w8_project_py
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import time
from typing import Union

from ...utils import timeunit, ITimeUnit, UnitOfTime
from .. import ITimer, RunStates

_time = time.time
IntOrFloat = Union[int, float]


class Timer(ITimer):

    def __init__(self, unit: UnitOfTime, wait_time: IntOrFloat):
        self._state: RunStates = RunStates.UN_START  # 状态
        self._step_time: IntOrFloat = 0  # 间隔时间
        self._start_time: IntOrFloat = 0  # 开始时间

        self._time_unit: ITimeUnit = timeunit(unit)  # 时间单位
        self._wait_time: IntOrFloat = wait_time  # 超时时间

    @property
    def is_started(self) -> bool:
        return self._state == RunStates.STARTED

    @property
    def is_timeout(self) -> bool:
        """
        is timeout
        :return:
        """
        return self._get_current_step() >= self._wait_time

    def current(self, **kwargs) -> IntOrFloat:
        """
        get current
        :param kwargs:
        :return:
        """
        return _time() * self._time_unit.get_time_unit()

    def start(self, **kwargs) -> None:
        if self.is_started:
            return

        # update state, start time
        self._set_state(RunStates.STARTED)
        self._set_start(self.current())

    def restart(self, **kwargs) -> None:
        """
        restart timer
        :return:
        """
        self._set_start(self.current())

    def stop(self, **kwargs) -> None:
        # update state, step time
        self._set_state(RunStates.UN_START)

    def clear(self) -> None:
        """
        clear timer
        :return:
        """
        self._set_state(RunStates.UN_START)
        self._set_start(0)

    def _set_start(self, start_time: IntOrFloat) -> None:
        self._start_time = abs(start_time)

    def _set_state(self, state: RunStates) -> None:
        self._state = state

    def _set_step(self, step: IntOrFloat) -> None:
        if step < 0:
            self._step_time = 0
        else:
            self._step_time = step

    def _get_current_step(self) -> IntOrFloat:
        """
        获取间隔时间
        :return:
        """
        return self.current() - self._start_time



# @Time     : 2021/6/1
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from typing import Callable, Union

from ..validators.base import AbstractValidator

ValidatorTypes = Union[AbstractValidator, Callable]


class IValidatorManager(object):

    def add(self, validator: ValidatorTypes) -> int:
        raise NotImplementedError("NotImplemented .add() -> int")

    def remove(self, validator: ValidatorTypes) -> int:
        raise NotImplementedError("NotImplemented .remove() -> int")

    def is_validate(self, value, **kwargs) -> bool:
        raise NotImplementedError("NotImplemented .is_validate() -> bool")



# @Time     : 2021/5/26
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .enums import EnumUtil
from .allowed import IAllowedStatic, Allowed
from .strings import Encoding, EncodingTypes, StringsUtil, StringOrBytesOrByteArray
from .path import PathUtil, PathTypes
from .time_unit import (
    ITimeUnit,
    SecondUnit,
    MilliSecondUnit,
    MicroSecondUnit,
    TimeUnit,
    UnitOfTime,
    timeunit,
    second,
    microsecond,
    millisecond
)



# @Time     : 2021/4/13
# @Project  : w8_project_py
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .base import (
    ICounter,
    ITimer,
    RunStates
)

from .core._ammeter import (
    AbstractMessageCounter,
    AsyncMessageCounter,
    MessageCounter,
)



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from enum import Enum
from typing import Dict, Union

from ..utils import EnumUtil
from ..validator.is_validators import is_dict, is_enum_subclass, is_none

AllowedOptionsTypes = Union[Dict, Enum]


class IAllowedStatic(object):
    """
    allowed static options interface, this is descriptors
    """

    def __get__(self, instance, owner):
        raise NotImplementedError("NotImplemented .__get__(interface, owner)")

    def add(self, name, allowed) -> None:
        raise NotImplementedError("NotImplemented .add(name, allowed) -> None")

    def get(self, allowed):
        raise NotImplementedError("NotImplemented .get(allowed)")

    def has(self, allowed) -> bool:
        raise NotImplementedError("NotImplemented .has(allowed)")


def _un_dict(options: Dict):
    if not options:
        return {}
    return {k: v for k, v in options.items()}


def _un_enum_subclass(options: Enum):
    return EnumUtil.unenum_to_dict(options)


class Allowed(IAllowedStatic):

    def __init__(self, options: AllowedOptionsTypes):
        self._options = self._to_options(options)

    def __get__(self, interface, owner):
        return self

    def add(self, name, option) -> None:
        self[name] = option

    def get(self, option):
        return self[option]

    def has(self, option):
        return option in self._options

    def __setitem__(self, name, allowed):
        if any([is_none(name), is_none(allowed)]):
            raise ValueError(f"allowed options key or value is not None")
        self._options[name] = allowed

    def __getitem__(self, option):
        if not self.has(option):
            raise KeyError(f"not found allowed {option}, from allowed {self}")
        return self._options[option]

    def __str__(self) -> str:
        return self._to_string()

    def _to_string(self, split: str = ", "):
        return split.join([str(member) for _, member in enumerate(self._options)])

    def _to_options(self, options: AllowedOptionsTypes) -> dict:
        if is_dict(options):
            return _un_dict(options)
        elif is_enum_subclass(options):
            return _un_enum_subclass(options)
        else:
            raise ValueError(
                f"options need Enum or dict, but got {type(options).__name__}"
            )



# @Time     : 2021/4/13
# @Project  : w8_project_py
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .. import ICounter


class Counter(ICounter):

    def __init__(self, count: int):
        self._count = 0 if not isinstance(count, int) else abs(count)

    def count(self, **kwargs) -> int:
        return self._count

    def increment(self, value: int, **kwargs) -> None:
        """
        update count
        :param value:
        :param kwargs:
        :return:
        """
        self._increment(value)
        if self._count < 0:
            self.clear()

    def clear(self, **kwargs) -> None:
        """
        clear count
        :param kwargs:
        :return:
        """
        self._count = 0

    def _increment(self, value: int) -> None:
        if not isinstance(value, int):
            value = 0

        self._count += value



# @Time     : 2021/5/26
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from pathlib import Path
from typing import AnyStr, List, Union

from .strings import StringsUtil
from ..validator.is_validators import is_any_string, is_list, is_path

PathTypes = Union[AnyStr, Path]


class PathUtil(object):

    @classmethod
    def get_path_from_base_dir(cls, basename: PathTypes, *paths) -> Path:
        base_dir = cls.get_base_dir(basename)
        if not base_dir.exists():
            raise ValueError(f"base dir need Path, but got {type(base_dir).__name__}")
        return base_dir.joinpath(*paths)

    @classmethod
    def get_base_dir(cls, basename: PathTypes) -> Path:
        """
        get base path dir
        :param basename:
        :return:
        """
        base_path: Path = cls.to_path(basename)
        if not base_path.exists():
            return Path.cwd()

        if base_path.is_file():
            return base_path.cwd()

        return base_path

    @classmethod
    def is_endswith_ext(cls, path: PathTypes, ext_names: List[str] = None) -> bool:
        """
        is file by ext name
        :param path:
        :param ext_names:
        :return:
        """
        file = cls.to_path(path)
        if not file.exists():
            raise ValueError(f"not found file from {path}")

        extensions = []
        if cls.is_extension_names(ext_names):
            extensions = ext_names
        if not extensions:
            return True
        return file.suffix in extensions

    @staticmethod
    def is_extension_names(value):
        return is_list(value)

    @classmethod
    def to_absolute(cls, path: PathTypes) -> Path:
        """
        to absolute path
        :param path:
        :return:
        """
        return cls.to_path(path).absolute()

    @classmethod
    def to_resolve(cls, path: PathTypes) -> Path:
        """
        to resolve path
        :param path:
        :return:
        """
        return cls.to_path(path).resolve()

    @classmethod
    def to_path(cls, anystr_or_path: PathTypes) -> Path:
        """
        string or bytes to path
        :param anystr_or_path:
        :return:
        """
        path: Path = None
        if is_path(anystr_or_path):
            path = anystr_or_path
        elif is_any_string(anystr_or_path):
            path = Path(StringsUtil.anystr_to_string(anystr_or_path))
        else:
            raise ValueError(
                f"need a string, bytes or Path, but got {type(anystr_or_path).__name__}"
            )
        return path



# @Time     : 2021/5/28
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
try:
    import simplejson as json
except ImportError:
    import json
from collections import defaultdict
from configparser import ConfigParser
from typing import Dict, DefaultDict, List, Union
from uuid import uuid4

from .base import AbstractConfReader
from .utils import ReaderUtil
from ..utils import EncodingTypes, PathTypes
from ..validator.is_validators import (
    is_dict,
    is_none,
    is_list,
    is_string
)


class IniReader(AbstractConfReader):
    DEFAULT = "DEFAULT"

    def __init__(self, file: PathTypes, encoding: EncodingTypes = None):
        super().__init__(file, encoding)
        self._parser = self._init_parser(self.file, self.encoding)

    def get_float(self, node: str, option: str) -> float:
        node = self._get_node_from_section(node)
        result = self._parser.getfloat(node, self._get_option_from_section(node, option))
        return result

    def get_int(self, node: str, option: str) -> int:
        node = self._get_node_from_section(node)
        result = self._parser.getint(node, self._get_option_from_section(node, option))
        return result

    def get_string(self, node: str, option: str) -> str:
        node = self._get_node_from_section(node)
        result = self._parser.get(node, self._get_option_from_section(node, option))
        return self._strip_quota(result)

    def has_node(self, node: str):
        if is_none(node):
            return False
        return self._parser.has_section(node)

    def _get_option_from_section(self, node: str, option: str):
        result = option
        if not self._parser.has_option(node, result):
            raise ValueError("not found option {option}")
        return result

    def _get_node_from_section(self, node: str):
        return IniReader.DEFAULT if not self.has_node(node) else node

    def _strip_quota(self, result: str) -> str:
        if not is_string(result):
            raise ValueError(f"result need string, but got {type(result).__name__}")
        if not result:
            return result
        return result.strip("'").strip('"')

    def _init_parser(self, file: PathTypes, encoding: str = None) -> ConfigParser:
        parser = ConfigParser()
        self._read_ini(parser, file, encoding)
        return parser

    def _read_ini(self,
                  parser: ConfigParser,
                  file: PathTypes,
                  encoding: str = None):
        parser.read(file, encoding)


Number = Union[float, int]
JsonTypes = Union[bool, Number, Dict, List, None]


class JsonNode(object):

    def __init__(self, key: str, data: JsonTypes = None):
        self._key = key
        self._data = self._create_by_json(key, data)

    def get(self, option: str, default=None):
        return self._data.get(option, default)

    def has_option(self, option):
        return option in self._data

    def _create_by_json(self, key: str, data: JsonTypes) -> dict:
        if is_dict(data):
            return self._by_dict(key, data)
        elif is_list(data):
            return self._by_list(key, data)
        else:
            return self._by_value(key, data)

    def _by_dict(self, key, data: Dict):
        return {k: v for k, v in data.items()}

    def _by_list(self, key, data: List):
        return {index: value for index, value in enumerate(data)}

    def _by_value(self, key, data: Union[bool, None, Number]):
        return {key: data}


class JsonReader(AbstractConfReader):

    def __init__(self, file: PathTypes, encoding: EncodingTypes = None):
        super().__init__(file, encoding)
        self._parser = self._init_parser(self.file, self.encoding)

    def has_node(self, node: str) -> bool:
        return node in self._parser

    def get_float(self, node: str, option: str) -> float:
        return self._get_from_parser(node, option, ReaderUtil.to_float)

    def get_int(self, node: str, option: str) -> int:
        return self._get_from_parser(node, option, ReaderUtil.to_int)

    def get_string(self, node: str, option: str) -> str:
        return self._get_from_parser(node, option, ReaderUtil.to_string)

    def _get_from_parser(self, node, option: str, callback=None):
        json_node = self._get_node_from_parser(node)
        result = self._get_option_from_parser(json_node, option)
        if is_none(callback):
            return result
        return callback(result)

    def _get_option_from_parser(self, node: JsonNode, option: str):
        if not node.has_option(option):
            raise KeyError(f"not found option {option}")
        return node.get(option)

    def _get_node_from_parser(self, node):
        return self._parser[node]

    def _init_parser(self, file: PathTypes, encoding: str = None) -> DefaultDict[str, JsonNode]:
        data = self._read_json(file, encoding)
        filtered = self._filter_json(data)
        return defaultdict(self._default_node, filtered)

    def _filter_json(self, result: dict) -> Dict[str, JsonNode]:
        if not is_dict(result):
            return {}
        return {key: self._default_node(key, value) for key, value in result.items()}

    def _read_json(self, file, encoding: str = None) -> dict:
        result = {}
        with open(file, mode="r") as f:
            result = json.load(f, encoding=encoding)
        return result

    def _default_node(self, key: str = None, value=None):
        _key: str = key
        if is_none(_key):
            _key = uuid4()
        return JsonNode(key, value)



# @Time     : 2021/6/3
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from enum import Enum

from .enums import EnumUtil

BASE = 10


class UnitOfTime(Enum):
    SECOND = BASE ** 0
    MILLISECOND = BASE ** 3
    MICROSECOND = BASE ** 6


class ITimeUnit(object):

    def get_time_unit(self) -> int:
        raise NotImplementedError()


class SecondUnit(ITimeUnit):

    def get_time_unit(self) -> int:
        return EnumUtil.unenum(UnitOfTime.SECOND)


class MilliSecondUnit(ITimeUnit):

    def get_time_unit(self) -> int:
        return EnumUtil.unenum(UnitOfTime.MILLISECOND)


class MicroSecondUnit(ITimeUnit):

    def get_time_unit(self) -> int:
        return EnumUtil.unenum(UnitOfTime.MICROSECOND)


class TimeUnit(ITimeUnit):

    def __init__(self, unit_of_time: UnitOfTime):
        self._unit = self._create_unit(unit_of_time)

    def get_time_unit(self) -> int:
        return self._unit.get_time_unit()

    def _create_unit(self, unit_of_time: UnitOfTime) -> ITimeUnit:
        if unit_of_time == UnitOfTime.SECOND:
            return SecondUnit()
        elif unit_of_time == UnitOfTime.MILLISECOND:
            return MilliSecondUnit()
        return MicroSecondUnit()


def timeunit(unit: UnitOfTime):
    return TimeUnit(unit).get_time_unit()


def second():
    """
    second
    :return:
    """
    return timeunit(UnitOfTime.SECOND)


def microsecond():
    """
    microsecond
    :return:
    """
    return timeunit(UnitOfTime.MICROSECOND)


def millisecond():
    """
    millisecond
    :return:
    """
    return timeunit(UnitOfTime.MILLISECOND)



# @Time     : 2021/4/13
# @Project  : w8_project_py
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import enum
import typing

T = typing.TypeVar("T")


class TimeoutStates(enum.Enum):
    """
    超时状态枚举
    """
    UN_TIMEOUT = 0
    TIMEOUT = 1


class RunStates(enum.Enum):
    """
    运行状态
    """
    UN_START = 0  # 停
    STARTED = 1  # 已开始
    PAUSE = 2  # 暂停


class ITimer(object):
    """
    消息计时器
    """

    @property
    def is_timeout(self) -> bool:
        """
        is_timeout
        :return:
        """
        raise NotImplementedError()

    def current(self, **kwargs) -> T:
        """
        get current time
        :param kwargs:
        :return:
        """
        raise NotImplementedError()

    def start(self, **kwargs) -> None:
        """
        start timer
        :param kwargs:
        :return:
        """
        raise NotImplementedError()

    def restart(self, **kwargs) -> None:
        raise NotImplementedError()

    def stop(self, **kwargs) -> None:
        """
        stop timer
        :param kwargs:
        :return:
        """
        raise NotImplementedError()

    def clear(self) -> None:
        """
        clear timeout
        :return:
        """
        raise NotImplementedError()


class ICounter(object):
    """
    消息计数器
    """

    def count(self, **kwargs) -> T:
        """
        get count
        :param kwargs:
        :return:
        """
        raise NotImplementedError()

    def increment(self, value: T, **kwargs) -> None:
        """
        increment
        :param value:
        :param kwargs:
        :return:
        """
        raise NotImplementedError()

    def clear(self, **kwargs) -> None:
        """
        clear
        :return:
        """
        raise NotImplementedError()



# @Time     : 2021/5/28
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .src.validator import is_validators
from .src.validator.error.base import BaseError
from .src.validator.error.errors import NotFunctionError, NotAsyncFunctionError
from .src.validator.validators.base import AbstractValidator
from .src.validator.validators.func import FunctionValidator
from .src.validator.validators.afunc import AsyncFunctionValidator
from .src.validator.manager.base import IValidatorManager, ValidatorTypes
from .src.validator.manager.managers import ValidatorManager, ValidatorList
from .src.validator.api import (
    is_validator_subclass,
    create,
    checker,
    check_function,
    check_async_function
)

from .src.utils.allowed import Allowed
from .src.utils.enums import EnumUtil
from .src.utils.path import PathUtil, PathTypes
from .src.utils.strings import StringsUtil, Encoding, EncodingTypes, StringOrBytesOrByteArray
from .src.utils.time_unit import (
    UnitOfTime,
    TimeUnit,
    SecondUnit,
    MilliSecondUnit,
    MicroSecondUnit,
    timeunit,
    second,
    microsecond,
    millisecond
)

from .src.callback.base import AbstractCallbackManager
from .src.callback.manager import CallbackManager, AsyncCallbackManager

from .src.conf.base import AbstractConfReader, IReaderFactory
from .src.conf.reader import IniReader, JsonReader
from .src.conf.factory import ReaderFactory

from .src.policy.base import AbstractExecutor, CoroOrFunction, ArgsTypes, KwargsTypes, ReturnType
from .src.policy.executor import Executor
from .src.policy.task import TaskFactory

from .src.ammeter import AbstractMessageCounter, AsyncMessageCounter, MessageCounter



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from pathlib import Path


def is_path(value):
    return isinstance(value, Path)



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from functools import partial

from .validators import AbstractValidator
from .validators.func import FunctionValidator
from .validators.afunc import AsyncFunctionValidator


def is_validator_subclass(klass):
    return issubclass(klass, AbstractValidator)


def _factory(subclass: AbstractValidator, message: str):
    if not is_validator_subclass(subclass):
        raise ValueError(f"need a {AbstractValidator.__name__} subclass, but got {type(subclass).__name__}")
    return partial(subclass, message)


def create(subclass: AbstractValidator, message: str, **kwargs) -> AbstractValidator:
    f = _factory(subclass, message)
    return f(**kwargs)


def checker(
        subclass: AbstractValidator,
        value,
        message: str,
        **kwargs
):
    validator = create(subclass, message)
    return validator(value)


def check_function(value):
    return checker(
        FunctionValidator,
        value,
        f"value need function or Callable, but got {type(value).__name__}"
    )


def check_async_function(value):
    return checker(
        AsyncFunctionValidator,
        value,
        f"value need async function, but got {type(value).__name__}"
    )



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .utils import ReaderUtil
from .base import AbstractConfReader, IReaderFactory
from .reader import IniReader, JsonReader, JsonNode
from .factory import ReaderFactory



# @Time     : 2021/5/26
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from scripts.build import build_package

name = "f1z1_common"
version = "0.3.12"

if __name__ == '__main__':
    build_package(name, version, filename="README.md")


# @Time     : 2021/6/1
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .base import AbstractValidator
from ..error import NotAsyncFunctionError
from ..is_validators import is_async_function


class AsyncFunctionValidator(AbstractValidator):

    def is_validate(self, value, **kwargs) -> bool:
        return is_async_function(value)

    def raise_error(self, message: str):
        raise NotAsyncFunctionError(message)



# @Time     : 2021/5/31
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .base import IReaderFactory
from .reader import IniReader, JsonReader
from ..utils import PathUtil


class ReaderFactory(IReaderFactory):

    @classmethod
    def create(cls, file, encoding=None, **kwargs):
        if cls.is_ini(file):
            return IniReader(file, encoding)
        elif cls.is_json(file):
            return JsonReader(file, encoding)
        else:
            raise ValueError(f"file ext name error, {file}")

    @staticmethod
    def is_json(file):
        return PathUtil.is_endswith_ext(file, [".json"])

    @staticmethod
    def is_ini(file):
        return PathUtil.is_endswith_ext(file, [".ini"])



# @Time     : 2021/6/1
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from .base import AbstractValidator
from .func import FunctionValidator
from .afunc import AsyncFunctionValidator



# @Time     : 2021/5/30
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
"""
conf module interfaces
"""
from abc import ABCMeta, abstractmethod

from ..utils import EncodingTypes, PathTypes
from . import ReaderUtil


class AbstractConfReader(metaclass=ABCMeta):
    """
    config reader abc class
    """

    def __init__(self, file: PathTypes, encoding: EncodingTypes = None):
        self._file = file
        self._encoding = ReaderUtil.to_encoding(encoding)

    @property
    def file(self):
        return self._file

    @property
    def encoding(self):
        return self._encoding

    @abstractmethod
    def has_node(self, node) -> bool:
        raise NotImplementedError("NotImplemented .has_node(node) -> bool")

    @abstractmethod
    def get_int(self, node: str, option) -> int:
        raise NotImplementedError("NotImplemented .get_int(node, option) -> int")

    @abstractmethod
    def get_float(self, node: str, option) -> float:
        raise NotImplementedError("NotImplemented .get_float(node, option) -> float")

    @abstractmethod
    def get_string(self, node: str, option) -> str:
        raise NotImplementedError("NotImplemented .get_string(node, option) -> str")


class IReaderFactory(object):

    @classmethod
    def create(cls, file: PathTypes, encoding: EncodingTypes, **kwargs) -> AbstractConfReader:
        raise NotImplementedError("NotImplemented .create() -> AbstractConfReader")



# @Time     : 2021/5/27
# @Project  : f1z1-g
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
from asyncio import iscoroutine, iscoroutinefunction
from collections.abc import Callable


def is_coroutine(value):
    return iscoroutine(value)


def is_function(value):
    return isinstance(value, Callable)


def is_async_function(value):
    return all([
        is_function(value),
        iscoroutinefunction(value)
    ])



# @Time     : 2021/4/14
# @Project  : w8_project_py
# @IDE      : PyCharm
# @Author   : Angel
# @Email    : 376355670@qq.com
import abc
from functools import wraps
from typing import Callable

from ...utils import UnitOfTime
from ...validator import check_function, is_function

from . import (
    Counter,
    Timer,
    IntOrFloat
)

Function = Callable


class AbstractMessageCounter(metaclass=abc.ABCMeta):
    """
    消息流
    """

    def __init__(self,
                 callback: Function,
                 wait_time: IntOrFloat,
                 time_unit: UnitOfTime = UnitOfTime.MILLISECOND):
        self._counter = Counter(0)
        self._timer = Timer(time_unit, wait_time)

        check_function(callback)
        self._callback = callback

    @property
    def callback(self) -> Function:
        """
        timeout callback
        :return:
        """
        return self._callback

    @abc.abstractmethod
    def tracking(self, func: Function) -> Function:
        raise NotImplementedError()

    @property
    def is_timeout(self):
        return self._timer.is_timeout

    def auto_increment(self, value: int) -> None:
        """
        auto increment
        :param value:
        :return:
        """
        self._counter.increment(value)

    def start(self) -> None:
        self._timer.start()

    def restart(self) -> None:
        """
        restart
        :return:
        """
        self._timer.restart()
        self._counter.clear()

    def trigger(self) -> None:
        cb = self._callback
        if is_function(cb):
            cb(self._counter.count())


class MessageCounter(AbstractMessageCounter):
    """
    同步消息计流器
    """

    def __init__(self,
                 callback: Function,
                 wait_time: IntOrFloat,
                 time_unit: UnitOfTime = UnitOfTime.MILLISECOND):
        super().__init__(callback, wait_time, time_unit)

    def tracking(self, func: Function) -> Function:
        """
        sync tracking
        :param func:
        :return:
        """
        flower = self

        @wraps(func)
        def decorate(*args, **kwargs):
            flower.start()
            result = func(*args, **kwargs)
            flower.auto_increment(1)
            flower.trigger()
            if flower.is_timeout:
                flower.restart()
            return result

        return decorate


class AsyncMessageCounter(AbstractMessageCounter):
    """
    异步消息计流器
    """

    def __init__(self,
                 callback: Function,
                 wait_time: IntOrFloat,
                 time_unit: UnitOfTime = UnitOfTime.MILLISECOND):
        super().__init__(callback, wait_time, time_unit)

    def tracking(self, func: Function) -> Function:
        """
        异步 tracking
        :param func:
        :return:
        """
        flower = self

        @wraps(func)
        async def async_decorate(*args, **kwargs):
            flower.start()
            result = await func(*args, **kwargs)
            flower.auto_increment(1)
            flower.trigger()
            if flower.is_timeout:
                flower.restart()
            return result

        return async_decorate
